.TL
Type System Interface
.SH
Purpose
.PP
This type system interface exists to abstract type system
functionality for the \fBOQL\fP to \fBAqua\fP translator and type
checker.
This allows the translator to be used with different OQL/Aqua 
compatible type systems.
With this isolation, the only translator component which should 
be need to be reimplemented is this type system interface.
.SH
Least Common Supertype
.PP
The \fILeast Common Supertype\fP (LCS) is a
property of two classes;
The LCS is the supertype common to
both classes which is farthest
from the root of the class tree,
or closest to both of the classes.

In the type system, the LCS concept is also used
with non-object types; it is also used to determine
compare-ability and arithmetic-ability of 
types (such as type \fIinteger\fP and \fIreal\fP can
be added to each other).

The LCS is also used to determine the common
subset of \fItuple\fP (and possibly \fIstruct\fP) members.

The LCS of two collections is a collection of the LCS type 
the the collected types.
.SH
OQL Types
.PP
This next table is a list of the types that OQL
uses, and the categories which I have placed them
into.
It is also the type tree.
.TS
box, center;
l s
l l.
Special Types
_
Void
Error
Nil
=
.T&
l s s
l l l.
Atomic ("Primitive") Types
_
Integer		PrimitiveType
Real
Boolean
Char
String
=
.T&
l s s
l l l.
ContainerType
_
Struct	<common struct>	StructType
Tuple	Name + any type allowed	TupleType
Object	<full blown object>	ObjectType
=
.T&
l s s
l l l.
CollectionType
_
Bag	<elements>	BagType
Set	<unique elements>	SetType
=
.T&
l s s
l l l.
IndexedCollectionType
_
Array		ArrayType
List		ListType
=
.T&
l l l.
Reference		RefType
=
Method		MethodType
.TE
.bp
.SH
class Type
.PP
.TS
allbox, center;
l lw(3i).
char *name()	String rep of name of type
kind()	?? enum TypeKinds
size()	size of type object
inline_size()	size of type object in an object
align()	alignment of type object (inline)
=
.T&
l s
l lw(3i).
Downcasts
_
ContainerType *isContainer()
TupleType *isTuple()
ObjectType *isObject()
StructType *isStruct()
MethodType *isMethod()
CollectionType *isCollection()
IndexedCollectionType *isIndexedCollection()
ArrayType *isArray()
ListType *isList()
BagType *isBag()
SetType *isSet()
int isInteger()
int isBoolean()	
=
Type *canCompareWith(Type *)	T{
If the type can be compared with the
the argument type, return a pointer to the
type of the comparison.
T}
Type *canDoMathWith(Type *)	T{
If this type can be arithmetically 
operated upon with the argument type,
return a pointer to the type
of the arithmetic operation.
T}
Type *canDoMath()	T{
If arithmetic operations can be performed
on this type, return the Type of the
result.
T}
=
Type *lcs(Type *)	T{
If this type and the argument type have a
\fILCS\fP, return the type of the LCS.
See the LCS section for details.
T}
.TE
.PP
The type system also maintains
static types, \fBTypeVoid\fP, \fBTypeError\fP,
and \fBTypeNil\fP.
.SH
class PrimitiveType
.PP
Primitive types are those types which
implement storage of values.
There are no subtypes of primitive types.
In the future, it may be desirable to have multiple
implementations of each primitive type, for example
float .vs. double, short .vs. long, range constraints
on an integer; but nothing like that exists now.
.PP
The type checker assumes the presence
of the following named atomic types.
.TS
allbox,center;
l.
integer
real
char
string
boolean
.TE
.LP
.TS
allbox, center;
l lw(3i).
void extract(Descriptor &d, int offset, void *where)	T{
Extract the value of the primitive type at d(offset) to
where.
T}
.TE
.PP
Currently, all these primitive types are implemented 
as the class PrimitiveType.
Seperating these types into their own classes is possible,
and probably desireable :-).
.SH
class ContainerType
.PP
A container is a class which contains a set of
named members.
Examples of containers are objects, structs, and tuples.
.LP
.TS
allbox, center;
l lw(3i).
Type *member(char *name)	T{
If 'name' is a member of the
container, return the type of
that member, else return NULL.
T}
ContainerType_i *iterator()	T{
Return a pointer to a container iterator.
This iterator will allow you to find
the name, type, and offset of each member
of the container.
T}
.TE
.SH 
class ObjectType
.PP
An object is a class which may inherit from another object.
.LP
.TS
allbox, center;
l lw(3i).
int isa(ObjectType *)	T{
True if the type has the
argument type as one of its
supertypes.
XXX ??? equivalent to type->lcs(t) == t?
T}
.TE
.SH
class TupleType
.PP
Tuples are short duration structures used for query
evaluation.  
At this point they are the only types which can be modified
during run time.
Tuples are not pushed back to the store, so you can't
store objects of type tuple.
.LP
.TS
allbox,center;
l lw(3i).
void concat(TupleType &)	T{
Concatenate all members of the argument
type to this type.
Used for tuple concatentation.
T}	
int add(char *name, Type *type)	T{
Add a member of 'name', 'type' to the tuple.
T}
.TE
.SH
class MethodType
.PP
Methods are function specifications,
which specify a type for return and arguments.
As OQL allow methods to take named arguments,
that is all the interface allows.
To implement positional arguments, just give
them a name.
For example, the aqua compiler uses '[%d]' for
positional argument specifications.
Mixing positional and named arguments
is a pain.
Using the argument 'name' for everything keeps
this problem out of this interface.
.LP
.TS
allbox,center;
l lw(3i).
Type *returns()	T{
The type of the method's return value.
T}
Type *argument(char *name)	T{
If an argument with 'name' exists, return
its type, otherwise NULL.
If the argument is a positional argument,
its name looks like '[%d]'.
T}
.TE
.SH
class CollectionType
.PP
Collections are a collection of objects of 
the same type.
There are two primary subclasses of Collections,
\fBIndexedCollections\fP, and \fBBags\fP.
An \fBIndexedCollection\fP allows for ordered access
to the collection (it could be called OrderedCollection
instead of IndexedCollection).
As such, it implements Lists and Arrays.

The other subclass, \fBBag\fP, is for unordered collections
of objects.
A subclass of Bag, \fBSet\fP, is for unordered collections
of objects that don't allow duplicates.
.LP
.TS
allbox,center;
l lw(3i).
Type *of()	T{
Return the type that the
collection is of.
T}
Type *rebuild(Type *of)	T{
Build a new collection of 'of's.
The new collection
is of the same kind of collection as
the type object.
T}
CollectionAccessor *accessor(Descriptor &d, int offset)	T{
Return a 'new' accessor the collection at the
d.offset location.
T}
.TE
.SH
class TypeDB
.PP
This class is used to interface with the system
type catalogs.
.TS
center,allbox;
l lw(3i).
Type *lookup(char *name)	T{
Look for type 'name' in the catalogs.
Return a pointer to it if it is found.
T}
Type *force_lookup(char *name)	T{
Look for type 'name' in the catalogs.
If it isn't found, return a pointer
to the 'TypeError'.
T}
.TE
.SH
class ExtentDB
.PP
This class is used to interface with the system
named object directory.
.TS
center, allbox;
l lw(3i).
Type *lookup(char *name)	T{
Look for the named object 'name', and
return its type, NULL if it doesn't
exist.
T}
.TE
.SH
class ContainerType_i
.PP
This class is used as an iterator to browse members
of containers.
.TS
center, allbox;
l lw(3i).
bool next(char *&name, Type *&type, int &offset)	T{
If any members of the container are left, return true, and
shove the membername, type and offset of the member
into the appropriate refs.
Otherwise return false.
T}
.TE
.SH
class CollectionAccessor
.PP
This class is the interface for accessing members
of a collection.
.TS
center, allbox;
l lw(3i).
int count()	T{
Number of items.
T}
Type &of()	T{
Type of collected objects.
T}
bool at(int position, Descriptor &d)	T{
If position is out of range, returns false.
Otherwise, returns true and sets the descriptor
to represent the position'thed item in the collection.
T}
bool next(Descriptor &d)	T{
If there is a next object to examine, return
true and set the descriptor to access it.
Otherwise return false.
T}
bool more()	T{
Are there more objects to examine?
T}
.TE
.SH
Notes on all of this
.PP
The \fIAqua\fP type checker assumes that it can generate
\fICollection\fPs on the fly, either thru the use of
\fBnew\fP or \fBrebuild()\fP.
.PP
IndexedCollectionTypes did have a \fIcardinality()\fP 
method, but the type checker doesn't generate the
values needed to range-check indices, so it isn't
mentioned here.
.PP
The various isXXX() methods can be used for
down-casting, or for boolean evaluations.
.PP
Memory allocation is in a real mess at this point.
Many of the functions *can* return new types,
or a pointer to an existing type.
Something has to be done to standardize and fix this,
but ???.
.PP
Type equality has not been implemented yet, partially
because I'm not sure of the semantics yet.
.PP
The \fBkind()\fP should be disposed of.
However, I first need to make a new method 
that checks two container or collection types
for container or collection compatibility.
??? LCS  can be used for this?
.PP
I've forgot to mention that all the above classes
have an \fBostream\fP friend '<<' output operator
attached to them.
