#ifndef _AQUA_H_
#define _AQUA_H_
/*
   Aqua abstract syntax tree objects
 */

/*
   Operator types

   lots
 */
#include "aqua_ops.h"

extern "C" char *strdup(const char *);

class TypeDB;
class ExtentDB;
class SymbolTable;
class oqlContext;
class typecheckState;

class aqua_t {
	friend ostream& operator<<(ostream &s, aqua_t &aqua_expr);
	/*
	   was protected:, but 2.6 doesn't allow members to access
	   protected fields of arguments of same class
	 */
public:
	Type	*_type;		// generated (or implied) data type
	aqua_op		_op;

	char	*common();

	// synthesize type info
	virtual	Type	*gen(typecheckState &) {
		return _type;
	}

	static	bool	print_types;
public:
	aqua_t() { _type = &TypeVoid; }
	virtual ~aqua_t() { }

	Type	*type() { return _type; }
	virtual	ostream &print(ostream &s, int level = 0);
	Type	*typecheck(oqlContext &context);
	aqua_op	op() { return _op; }
	static	void	setTypePrinting(bool onoff) { print_types = onoff; }
};

class aqua_error_t : public aqua_t {
public:
	aqua_error_t() { _type = &TypeError; }
	virtual	ostream &print(ostream &s, int level = 0);
};

class aqua_constant_t : public aqua_t {
public:
	auto_string	_rep;
	aqua_constant_t(Type *type, char *rep) : _rep(rep) {
		_type = type;
	}

	virtual	ostream &print(ostream &s, int level = 0);
//	Type	*gen(typecheckState &);
};

class aqua_type_t : public aqua_t {
public:
	auto_string	_name;
	aqua_type_t(Type *type, char *name) : _name(name) {
		_type = type;
	}
	aqua_type_t(char *name);

	virtual	ostream &print(ostream &s, int level = 0);
	Type	*gen(typecheckState &);
};

class aqua_dbvar_t : public aqua_t {
public:
	auto_string	_name;
	Type	*_extent;
	/* Symbol *symbol */
	aqua_dbvar_t(char *name, Type *extent) : _name(name) {
		_op = a_dbvar;
		_extent = extent;
	}
	
	virtual	ostream &print(ostream &s, int level = 0);
	Type	*gen(typecheckState &);
};

class aqua_var_t : public aqua_t {
public:
	auto_string	_name;
	/* Symbol *symbol */
	aqua_var_t(/*ignored*/ aqua_op op, char *name) : _name(name) {
		_op = a_var;
	}
	aqua_var_t(char *name) : _name(name) {
		_op = a_var;
	}
	
	virtual	ostream &print(ostream &s, int level = 0);
	Type *gen(typecheckState &);
};

class aqua_use_t : public aqua_t {
public:
	auto_string	_name;
	aqua_use_t(char *name) : _name(name) {
		_op = a_use;
	}
	
	virtual	ostream &print(ostream &s, int level = 0);
	Type *gen(typecheckState &);
};

class aqua_tuple_t : public aqua_t {
public:
	auto_string	_name;
	aqua_t	*_value;
	aqua_tuple_t(char *name, aqua_t *value) : _name(name) {
		_value = value;
	}

	virtual	ostream &print(ostream &s, int level = 0);
	Type *gen(typecheckState &);
};

class aqua_apply_t;

class aqua_function_t : public aqua_t {
public:
	aqua_var_t	*_one;
	aqua_var_t	*_two;
	aqua_t		*_body;
	aqua_function_t(aqua_var_t *one, aqua_t *body) {
		_op = a_function;
		_one = one;
		_two = (aqua_var_t *)0;
		_body = body;
	}
	aqua_function_t(aqua_var_t *one, aqua_var_t *two, aqua_t *body) {
		_op = a_function;
		_one = one;
		_two = two;
		_body = body;
	}
	ostream	&print(ostream &s, int level = 0);
	Type	*gen(typecheckState &,
		     Type *arg0 = (Type *)0, Type *arg1 = (Type *)0);
};

/* aqua_function_user_t :-) */
class aqua_apply_t : public aqua_t {
public:
	aqua_function_t	*_what;	/* apply this */
	aqua_t	*_to;		/* to this */

	aqua_apply_t(aqua_op op, aqua_function_t *what, aqua_t *to) {
		_op = op;
		_what = what;
		_to = to;
	}
	/* backwards compatibility */
	aqua_apply_t(aqua_op op, aqua_t *what, aqua_t *to) {
		_op = op;
		_what = (aqua_function_t *)what;
		_to = to;
	}

	ostream &print(ostream &s, int level = 0);
	Type	*gen(typecheckState &);
};

class aqua_join_t : public aqua_apply_t {
public:
	aqua_t	*_other;	/* other source of joined tuples */

	aqua_join_t(aqua_function_t *what, aqua_t *left, aqua_t *right) : aqua_apply_t(a_tuple_join, what, left) {
		_other = right;
	}
	ostream	&print(ostream &s, int level = 0);
	Type	*gen(typecheckState &);
};

class aqua_fold_t : public aqua_apply_t {
public:
	/* _what is treated as the extractor */
	aqua_t	*_initial;
	aqua_function_t	*_folder;
	aqua_fold_t(aqua_t *initial,
		    aqua_function_t *extractor,
		    aqua_function_t *folder,
		    aqua_t *source) : aqua_apply_t(a_fold, extractor, source) {
			    _folder = folder;
			    _initial = initial;
	}
	virtual	ostream &print(ostream &s, int level = 0);
	Type *gen(typecheckState &);
};


class aqua_trinary_t : public aqua_t {
public:
	aqua_t	*_left;
	aqua_t	*_middle;	
	aqua_t	*_right;
	aqua_trinary_t(aqua_op op,aqua_t *left,aqua_t *middle, aqua_t *right) {
		_op = op;
		_left = left;
		_middle = middle; 
		_right = right;
	}
	virtual	ostream &print(ostream &s, int level = 0);
	Type *gen(typecheckState &);
};

class aqua_binary_t : public aqua_t {
public:
	aqua_t	*_left;
	aqua_t	*_right;
	aqua_binary_t(aqua_op op, aqua_t *left, aqua_t *right) {
		_op = op;
		_left = left;
		_right = right;
	}
	virtual	ostream &print(ostream &s, int level = 0);
	virtual Type *gen(typecheckState &);
};

class aqua_unary_t : public aqua_t {
public:
	aqua_t	*_what;
	aqua_unary_t(aqua_op op, aqua_t *what) {
		_op = op;
		_what = what;
	}
	virtual	ostream &print(ostream &s, int level = 0);
	virtual Type *gen(typecheckState &);
};

class aqua_get_t : public aqua_unary_t {
public:
	auto_string	_name;
	aqua_get_t(aqua_t *what, char *name)
		: aqua_unary_t(a_get, what), _name(name) {
	}
	
	virtual	ostream &print(ostream &s, int level = 0);
	virtual	Type	*gen(typecheckState &);
};

/* XXX consistency problems */
class aqua_method_t : public aqua_get_t {
public:
	aqua_t	*_args;
	aqua_method_t(aqua_t *what, char *method, aqua_t *args)
		: aqua_get_t(what, method) {
			_op = a_method;
			_args = args;
	};
	virtual ostream &print(ostream &s, int level = 0);
	Type	*gen(typecheckState &);
};

class aqua_assign_t : public aqua_binary_t {
public:
	aqua_assign_t(aqua_t *to, aqua_t *what)
		: aqua_binary_t(a_assign, to, what) { }
	// virtual	ostream &print(ostream &s, int level = 0);
	Type	*gen(typecheckState &);
};
	

class aqua_zary_t : public aqua_t {
public:
	aqua_t	*_what;
	aqua_zary_t(aqua_op op) {
		_op = op;
	}
	virtual	ostream &print(ostream &s, int level = 0);
	Type	*gen(typecheckState &);
};


#endif /* _AQUA_H_ */ 
