#include <stdio.h>
#include <ostream.h>

#include "symbol.hh"
#include "types.hh"
#include "aqua.hh"
#include "typedb.hh"
#include "oql_context.hh"

#include "sstack.hh"

class typecheckState {
public:
	oqlContext	&oql;
	SymbolStack	current;
	typecheckState(oqlContext &_oql)
		: oql(_oql), current("typecheck symbol stack") { }

	TypeDB		&types() { return oql.db.types; }
	ExtentDB	&extents()  { return oql.db.extents; }
	SymbolTable	&defines()  { return oql.defines; }
};
	

/* Emacs indent mode screws up with {} case stmts */
#define	BEGIN	{/*void*/
#define	END	} break


Type *aqua_t::typecheck(oqlContext &context)
{
	typecheckState state(context);
	return gen(state);
}

Type *aqua_type_t::gen(typecheckState &state)
{
	return _type = state.types().force_lookup(_name);
}

Type *aqua_dbvar_t::gen(typecheckState &state)
{
	return _type = _extent;
}

Type	*aqua_var_t::gen(typecheckState &state)
{
	Symbol	*sym;
	sym = state.current.lookup_all(_name);
	if (!sym) {
		cerr << "aqua_var_t(" << _name << ") not found!\n";
		return &TypeError;
	}
	_type = (Type *)(sym->data());
	// cout << "Var(" << _name << ") : " << *_type << '\n';
	return _type;
}

Type *aqua_get_t::gen(typecheckState &state)
{
	ContainerType *it;
	RefType *ref;
	MethodType *method;

	aqua_unary_t::gen(state);

	/* assumes only 1 ref */
	ref = _type->isRef();
	if (ref)
		_type = ref->of();

	it = _type->isContainer();
	if (!it)
		return _type = &TypeError;
	_type = it->member(_name);
	if (!_type)
		_type = &TypeError;

#if 1
	/* XXX rewrite of node needed */
	if ((method = _type->isMethod())  &&  _op == a_get) {
		cerr << "get() -> method() rewrite needed!\n";
		/* XXX should check that method is Method(void) */
		_type = method->returns();
	}
#endif

	return _type;
}

Type *aqua_method_t::gen(typecheckState &state)
{
	MethodType *it;

	it = (aqua_get_t::gen(state))->isMethod();
	if (!it)
		return _type = &TypeError;

	/* XXX check arg types */

	return _type = it->returns();
}

Type	*aqua_function_t::gen(typecheckState &state, Type *arg0, Type *arg1)
{
	Type	*result;

	if (arg0 && !_one  ||  !arg0 && _one)
		return &TypeError;
	if (arg1 && !_two  ||  !arg1 && _two)
		return &TypeError;
#if 0
	/* This is good enough if we use the duplicated aqua_var_t */
	if (arg0)
		_one->_type = arg0;
	if (arg1)
		_two->_type = arg1;
#else
	/*
	   We must set both the embedded type and the symbol stack
	   types, for both types of vars are put in the tree
	   by aqua_tran (For now at least)

	   BZZT wrong; only need the symbol stack entries.
	 */  
	state.current.push("function");
	if (arg0) {
		_one->_type = arg0;
		state.current.enter(_one->_name, (char *)arg0);
		// cout << "Function(" << _one->_name << ", " << *arg0 << ")\n";  
	}
	if (arg1) {
		_two->_type = arg1;
		state.current.enter(_two->_name, (char *)arg1);
	}
#endif
	result = _body->gen(state);
#if 1
	state.current.pop();
#endif
	return _type = result;
}

Type	*aqua_use_t::gen(typecheckState &state)
{
	Symbol *sym;
	aqua_t *define;

	sym = state.defines().lookup(_name);
	if (!sym)
		return _type = &TypeError;
	define = (aqua_t *) sym->data();
	
	return define->type();
}

/* XXX ick what about double-variable functions ??? */
Type	*aqua_apply_t::gen(typecheckState &state)
{
	Type	*element;
	Type	*input;
	CollectionType *work_on;
	Type	*results_in;
	TupleType *group;

	input = _to->gen(state);
	/* major league conversion needed here from collection->object, etc */

	work_on = input->isCollection();
	if (!work_on) {
		// cout << "Apply must be to a collection\n";
		return _type = &TypeError;
	}
	element = work_on->of();
	
	// cout << "Apply working on " << *element << '\n';
	results_in = _what->gen(state, element);

	switch (_op) {
	case a_select:
		/* select just filters the input elements */
		if (results_in->isBoolean())
			results_in = input;
		else
			results_in = &TypeError;
		break;
	case a_apply:
		/* Sets turn into bags when an apply happens  */
		if (work_on->isSet())
			results_in = new BagType(results_in);
		else
			results_in = work_on->rebuild(results_in);
		break;
	case a_group:
		/* group produces set<struct<"attr":x, "group":set<input>>> */
		group = new TupleType();
		group->add("attr", results_in);
		group->add("group", new SetType(element));
		results_in = new SetType(group);
		break;
	case a_exists:
	case a_forall:
		if (!results_in->isBoolean())
			results_in = &TypeError;
		break;
	}

#if 0
	/* Can't guarantee a set remains that way */ 
	if (resuls_in->kind() == tt_set) {
		results_in = Type::_rebuild(tt_bag, results_in->of);
		delete results_in;
	}
#endif

	return _type = results_in;
}

Type	*aqua_join_t::gen(typecheckState &state)
{
	Type	*input1, *input2;
	CollectionType *tmp;
	TupleType	*tuple;
	Type	*t;
	

	input2 = _other->gen(state);
	input1 = _to->gen(state);

	/* convert inputs from collections to items */
	tmp = input1->isCollection();
	if (!tmp)
		return _type = &TypeError;
	input1 = tmp->of();
	if (!input1->isTuple())
		return _type = &TypeError;

	tmp = input2->isCollection();
	if (!tmp)
		return _type = &TypeError;
	input2 = tmp->of();
	if (!input2->isTuple())
		return _type = &TypeError;

	t = _what->gen(state, input1, input2);
	if (!t->isBoolean())
		return _type = &TypeError;

	tuple = new TupleType();
	tuple->concat(*input1->isTuple());
	tuple->concat(*input2->isTuple());

	/* Bag, List, ??? */ 
	t = new BagType(tuple);
	
	return _type = t;
}

Type *aqua_tuple_t::gen(typecheckState &state)
{
	Type *it = _value->gen(state);
	TupleType *tuple;

	_type = tuple = new TupleType();
	tuple->add(_name, it);

	return _type;
}

Type *aqua_assign_t::gen(typecheckState &state)
{
	Type *to = _left->gen(state);
	Type *from = _right->gen(state);

	_type = from;
	switch (_left->op()) {
	case a_dbvar:
		/* type incompatibility is ok; runtime changes dbvar type */
		_type = from;
	default:
		if (! to->lcs(from))
			_type = &TypeError;
		break;
	}
	return _type;
}

Type *aqua_zary_t::gen(typecheckState &state)
{
	switch (_op) {
	case a_true:
	case a_false:
		_type = state.types().force_lookup("boolean");
		break;
	case a_nil:
		_type = state.types().force_lookup("nil");
		break;
	case a_void:
		_type = state.types().force_lookup("void");
		break;
	}
	return _type;
}


Type *aqua_unary_t::gen(typecheckState &state)
{
	_type = _what->gen(state);

	switch (_op) {
	case a_abs:
	case a_negate:
		/* The type must be a numeric type */
		if (!_type->canDoMath())
			_type = &TypeError;
		break;
	case a_not:
		/* The type must be boolean */
		if (!_type->isBoolean())
			_type = &TypeError;
		break;
	case a_first:
	case a_last:
	case a_choose:
		/* the type must be a list, array, or collection */
		BEGIN;
		CollectionType *colxion = _type->isCollection();
		_type = colxion ? colxion->of() : &TypeError;
		END;
	case a_multiset_to_set:
		BEGIN;
		BagType *bag = _type->isBag();
		CollectionType *colxion;
#if 0
		_type = bag ? new SetType(bag->of()) : &TypeError;
#else
		if (bag)
			_type = new SetType(bag->of());
		else if (colxion = _type->isCollection()) {
			_type = new SetType(colxion->of());
			/* still needs rewrite */
		}
		else
			_type = &TypeError;
			
#endif
		END;
	case a_list_to_set:
		BEGIN;
		ListType *list = _type->isList();
		_type = list ? new SetType(list->of()) : &TypeError;
		END;

	case a_flatten:
		BEGIN;
		CollectionType *outer = _type->isCollection();
		CollectionType *inner = outer ? outer->of()->isCollection() :0;
		BagType *bag;
		IndexedCollectionType *ixc, *ixc1;
		if (inner  &&  outer) {
			if (bag = inner->isBag()) {
				cout << "Flatten xx(bag) -> bag\n";
				_type = bag;
				break;
			}
			if ((bag = outer->isBag()) &&
			    (ixc = inner->isIndexedCollection())) {
				cout << "Flatten set(list) -> set\n";
				_type = new SetType(ixc->of());
				break;
			}
			if ((ixc = outer->isIndexedCollection()) &&
			    (ixc1 = inner->isIndexedCollection())) {
				cout << "Flatten list(list) -> list\n";
				_type = ixc1;
				break;
			}
			cout << "Flatten ????\n";
			_type = &TypeError;
		}
		END;
		
	case a_array:
		_type = new ArrayType(_type);
		break;
	case a_list:
		_type = new ListType(_type);
		break;
	case a_set:
		_type = new SetType(_type);
		break;
	case a_multiset:
		_type = new BagType(_type);
		break;

	default:
		/* do nothing, just pass the type along */
		break;
	}
	return _type;
}

Type *aqua_binary_t::gen(typecheckState &state)
{
	Type *one, *two, *result;
	one = _left->gen(state);
	two = _right->gen(state);

	_type = &TypeError;	/* default value */
	switch (_op) {
	case a_tuple_concat:
		BEGIN;
		TupleType *l, *r;
		l = one->isTuple();
		r = two->isTuple();
		if (l && r) {
			TupleType *me;
			me = new TupleType;
			me->concat(*l);
			me->concat(*r);
			_type = me;
		}
		END;
	case a_array_concat:
	case a_list_concat:
		BEGIN;
		IndexedCollectionType *l, *r;
		Type	*lcs;
		l = one->isIndexedCollection();
		r = two->isIndexedCollection();
		if (l && r && (lcs = l->lcs(r)))
			_type = lcs;
		END;
	case a_union:
	case a_additive_union:
	case a_diff:
	case a_intersect:
		/*
		   unions resulting in a Bag should be additive
		   additive unions should be done on bags
	         */
		BEGIN;
		BagType *l, *r;
		Type	*lcs;
		l = one->isBag();
		r = two->isBag();
		if (l && r && (lcs = l->lcs(r)))
			_type = lcs;
		END;
	case a_eq:
	case a_ne:
		/*
		   Different from relative compares, as objects
		   can be compared for equality
		   XXX sort of disgusting
		 */  
		/* coercions may be needed */
		if (result = one->canCompareWith(two))
			_type = state.types().force_lookup("boolean");
		break;
	case a_le:
	case a_lt:
	case a_gt:
	case a_ge:
		/* coercions may be needed */
		/* XXX relative compare method needed */
		if (result = one->canDoMathWith(two))
			_type = state.types().force_lookup("boolean");
		break;
	case a_add:
	case a_subtract:
	case a_multiply:
	case a_divide:
	case a_max:
	case a_min:
		if (result = one->canDoMathWith(two))
			_type = result;
		else if (_op == a_add) {
			/* Is it really list/array concatenation ? */
			IndexedCollectionType *l, *r;
			Type	*lcs;
			l = one->isIndexedCollection();
			r = two->isIndexedCollection();
			if (l && r && (lcs = l->lcs(r))) {
				_type = lcs;
				/* disgusting, should new /delete stuff */
				_op = lcs->isArray()
					? a_array_concat
					: a_list_concat;	
			}
		}
		break;
	case a_mod:
		if (one->isInteger()  &&  two->isInteger())
			_type = state.types().force_lookup("integer");
		break;
	case a_and:
	case a_or:
		if (one->isBoolean()  &&  two->isBoolean())
			_type = state.types().force_lookup("boolean");
		break;
	case a_at:
		BEGIN;
		IndexedCollectionType *array = one->isIndexedCollection();
		int	indexOk = two->isInteger();
		if (array && indexOk)
			_type = array->of();
		END;
	case a_cast:
		/* This catches most things except ... */
		/* XXX and the bloody ref problem shows up again */
		if (one->lcs(two))
			_type = one;
		else {
			/* This catches down-classing */
			ObjectType *type = one->isObject();
			ObjectType *it = two->isObject();
			if (type && it && (type->isa(it)))
				_type = type;
			else
				cout << *two << " is not a " << *one << '\n';
		}
		break;
	case a_sort:
		BEGIN;
		ListType *keys = one->isList();
		CollectionType *source = two->isCollection();
		/* sort output is always a list */
		if (keys && source)
			_type = new ListType(source->of());
		END;
	case a_member:
		BEGIN;
		CollectionType *colxion = two->isCollection();
		if (colxion && one->lcs(colxion->of()))
			_type = state.types().force_lookup("boolean");
		END;
	default:
		cerr << "Unused binary " << *this << '\n';
		break;
	}

	return _type;
}

Type *aqua_trinary_t::gen(typecheckState &state)
{
	Type *one, *two, *three;
	IndexedCollectionType *array;

	/* there is only 1 trinary  type at this point */
	one = _left->gen(state);
	two = _middle->gen(state);
	three = _right->gen(state);

	/* range can return one or more elements, so type is same as src */
	array = one->isIndexedCollection();
	if (array  &&  two->isInteger()  &&  three->isInteger())
		_type = array;
	else
		_type = &TypeError;
	return _type;
}

Type *aqua_fold_t::gen(typecheckState &state)
{
	CollectionType *source;		/* source of whatevers */
	Type	*element;		/* individual whatever */
	Type	*extract;		/* extract from whatever */
	Type	*initial;		/* initial value */
	Type	*folded;		/* folded result */

	/* there is only 1 cubary  type at this point */
	_type = &TypeError;

	source = _to->gen(state)->isCollection();
	if (!source)
		return _type;

	element = source->of();
	extract = _what->gen(state, element);
	initial = _initial->gen(state);
	folded = _folder->gen(state, initial, extract);

	return _type = folded;
}
