#ifndef _TYPES_HH_
#define _TYPES_HH_

class PrimitiveType;
class ContainerType;
class TupleType;
class CollectionType;
class IndexedCollectionType;
class ObjectType;
class ListType;
class ArrayType;
class MethodType;
class SetType;
class BagType;
class RefType;

extern "C" char *strdup(char *);

class Type {
	friend	ostream &operator<<(ostream &s, Type &type);
protected:
	char		*_name;

public: /* protected: */
	Type(char *name) { _name = name; }

public:
//	int	isNamed() { return _name != 0; }
	virtual char	*name();

#if 0
{
	return _name ? _name : strdup(form("#%#x", (int)this));
}
#endif

	virtual	PrimitiveType	*isPrimitive() { return 0; }
	virtual	ContainerType	*isContainer() { return 0; }
	virtual	TupleType	*isTuple() { return 0; }
	virtual	ObjectType	*isObject() { return 0; }

	virtual	MethodType	*isMethod() { return 0; }

	virtual	CollectionType	*isCollection() { return 0; }
	virtual IndexedCollectionType	*isIndexedCollection() { return 0;}
	virtual	ArrayType	*isArray() { return 0; }
	virtual	ListType	*isList()  { return 0; }
	virtual SetType		*isSet() { return 0; }
	virtual BagType		*isBag() { return 0; }

	virtual RefType	*isRef() { return 0; }

	virtual	int	isInteger() { return 0; }
	virtual int	isBoolean() { return 0; }

	virtual Type	*canCompareWith(Type *it) { return 0; }
	virtual	Type	*canDoMathWith(Type *it) { return 0; }
	virtual Type	*canDoMath() { return 0; }
	
	virtual	Type	*lcs(Type *it) { return 0; }

	virtual ostream	&print(ostream &s) {
		return s << "<Type " << Type::name() << ">";
	}
	
};

class ErrorType : public Type {
public: /* protected: */
	ErrorType(char *name) : Type(name) {}
public:
};

class PrimitiveType : public Type {
public:
#include "primitive_types.h"
private:
	enum PrimitiveTypes	_primitive;
public: /*protected:*/
	PrimitiveType(char *name, PrimitiveTypes prim) : Type(name) {
		_primitive = prim;
	}
			
public:
	virtual	PrimitiveType	*isPrimitive() { return this; }
	enum PrimitiveTypes	primitive() { return _primitive; }
	Type	*lcs(Type *it);
	Type	*canCompareWith(Type *it);
	Type	*canDoMathWith(Type *it);
	Type	*canDoMath();
	int	isInteger() { return _primitive == pt_integer; }
	int	isBoolean() { return _primitive == pt_boolean; }
	// ostream	&print(ostream &s);
};


/*
   Base type for types containing named members
   This is really a reusable interface class for container objects.
 */

class ContainerType : public Type {
protected: /* private: */
	SymbolTable	members;
public: /* protected: */
	ContainerType(char *name) : Type(name) {}
	/* This is only visible for tuples in real life */
	int	add(char *name, Type *type) {
		Symbol *s = members.lookup(name);
		if (s)
			return 1;
		members.enter(name, (char *)type, 0);
		return 0;
	}
	ostream	&_print(ostream &s, char *typename);
public:
	virtual Type	*member(char *name) {
		Symbol	*s = members.lookup(name);
		return s ? (Type *)(s->data()) : (Type *)0;
	}
	virtual	ostream	&print(ostream &s) = 0;
	virtual	ContainerType	*isContainer() { return this; }
	virtual	Type	*canCompareWith(Type *of);
};


class ObjectType : public ContainerType {
private:
	ObjectType	*_superclass;
	void classlist(ObjectType **list, int &where, int remaining); 
public: /*protected:*/
	ObjectType(char *name, ObjectType *superclass = 0) : ContainerType(name) {
		_superclass = superclass;
			
	}
public:
	Type	*member(char *name) {
		Symbol *s;
		Type	*t;
		t = ContainerType::member(name);
		if (!t && _superclass)
			t = _superclass->member(name);
		return t;
	}
	int	isa(ObjectType *it);
	Type	*lcs(Type *it);
	virtual	ObjectType	*isObject() { return this; }
	ostream	&print(ostream &s) { return ContainerType::_print(s, 0);}
};


class StructType : public ContainerType {
private:
public: /* protected: */
	StructType(char *name) : ContainerType(name) {};
public:
	ostream	&print(ostream &s) { return ContainerType::_print(s, "Struct");}
};


class TupleType : public ContainerType {
private:
public: /* protected: */
	TupleType(char *name = 0) : ContainerType(name) {};
public:
	void concat(TupleType &from);
	char	*name();
	Type	*lcs(Type *it);
	virtual	TupleType	*isTuple() { return this; }
	ostream	&print(ostream &s) { return ContainerType::_print(s, "Tuple");}
};


class MethodType : public Type {
private:
	Type	*_returns;
	SymbolTable	args;

public: /*  protected: */
	MethodType(char *name, Type *returns) : Type(name) {
		_returns = returns;
	}
	MethodType(Type *returns) : Type(0) {
		_returns = returns;
	}
	int	addArg(char *name, Type *type) {
		Symbol *s = args.lookup(name);
		if (s)
			return 1;
		args.enter(name, (char *)type, 0);
		return 0;
	}

public:
	Type	*returns() { return _returns; }
	Type	*argument(char *name) {
		Symbol	*s = args.lookup(name);
		return s ? (Type *)(s->data()) : (Type *)0;
	}
	virtual	MethodType	*isMethod() { return this; }
	ostream	&print(ostream &s);
};


/*
   XXXX
   
   Reference share the interface that Collections use,
   but a reference isn't a collection, so it shouldn't be
   part of that class tree.
 */  

class RefType : public Type {
private:
	Type	*_of;
public: /* protected: */
	RefType(char *name, Type *of) : Type(name) {
		_of = of;
	}
	RefType(Type *of) : Type((char *)0) {
		_of = of;
	}
	void setOf(Type *of) { _of = of; }
public:
	Type	*of() { return _of; }
	virtual	RefType	*isRef() { return this; }
	Type	*canCompareWith(Type *of);
	Type	*rebuild(Type *of);

	Type	*lcs(Type *it);

	// Type	*operator->() { return _of; }
	Type *operator*() { return _of; }
	
	ostream	&print(ostream &s);
};

#if 0
class NilRefType : public RefType {
};
#endif

/*
   Collections of similar types

   As with the container, this class is a reusable interface class
 */

class CollectionType : public Type {
private:
	Type	*_of;
public: /* protected: */
	CollectionType(char *name, Type *of) : Type(name) {
		_of = of;
	}
	CollectionType(Type *of) : Type((char *)0) {
		_of = of;
	}
	void setOf(Type *of) { _of = of; }
	ostream	&_print(ostream &s, char *typename);
public:
	virtual	Type	*rebuild(Type *of) = 0;
	Type	*of() { return _of; }
	virtual	CollectionType	*isCollection() { return this; }

	virtual Type	*lcs(Type *it) = 0;
	Type	*_lcs(CollectionType *it);
	
	virtual ostream	&print(ostream &s) = 0;
};


#if 0
class RefType : public CollectionType {
public:
	RefType(char *name, Type *of) : CollectionType(name, of) {}
	RefType(Type *of) : CollectionType(of) {}
};
#endif

class BagType : public CollectionType {
public:
	virtual BagType		*isBag() { return this; }
	BagType(char *name, Type *of) : CollectionType(name, of) {}
	BagType(Type *of) : CollectionType(of) {}
	virtual	Type	*rebuild(Type *of) { return new BagType(of); }
	virtual Type	*lcs(Type *it);
	ostream	&print(ostream &s) { return CollectionType::_print(s, "Bag");}
};

class SetType : public BagType {
public:
	virtual SetType		*isSet() { return this; }
	SetType(char *name, Type *of) : BagType(name, of) {}
	SetType(Type *of) : BagType(of) {}
	virtual	Type	*rebuild(Type *of) { return new SetType(of); }
	Type	*lcs(Type *it);
	ostream	&print(ostream &s) { return CollectionType::_print(s, "Set");}
};

class IndexedCollectionType : public CollectionType {
	int	_cardinality;
public:
	IndexedCollectionType(char *name, Type *of) : CollectionType(name,of) {
		_cardinality = 0;
	}
	IndexedCollectionType(Type *of) : CollectionType(of) {
		_cardinality = 0;
	}
	virtual IndexedCollectionType	*isIndexedCollection() { return this;}
	int	cardinality() { return _cardinality; }
	void setCardinality(int newcard) { _cardinality = newcard; }
	virtual	Type	*rebuild(Type *of) = 0;
	virtual	Type	*lcs(Type *it) = 0;
	virtual	ostream	&print(ostream &s) = 0;
};

class ListType : public IndexedCollectionType {
public:
	virtual	ListType	*isList()  { return this; }
	ListType(char *name, Type *of) : IndexedCollectionType(name, of) {}
	ListType(Type *of) : IndexedCollectionType(of) {}
	virtual	Type	*rebuild(Type *of) { return new ListType(of); }
	Type	*lcs(Type *it);
	ostream	&print(ostream &s) { return CollectionType::_print(s, "List");}
};

class ArrayType : public IndexedCollectionType {
public:
	virtual	ArrayType	*isArray() { return this; }
	ArrayType(char *name, Type *of) : IndexedCollectionType(name, of) {}
	ArrayType(Type *of) : IndexedCollectionType(of) {}
	virtual	Type	*rebuild(Type *of) { return new ArrayType(of); }
	Type	*lcs(Type *it);
	ostream	&print(ostream &s) { return CollectionType::_print(s, "Array");}
};


extern PrimitiveType TypeVoid;
extern ErrorType TypeError;
extern ObjectType TypeNil; 

#endif /* _TYPES_HH_ */
